VERSION = '0.0.3'

ENDPOINT_AI = 'ai'
ENDPOINT_HH = 'hh'
COMMAND_GET_STATUS = 'getDeviceStatus'
COMMAND_GET_MODEL_DESC = 'getModelDescription'
COMMAND_GET_PROGRAM = 'getProgram'
COMMAND_GET_COMMAND = 'getCommand'
